/*------------------------------------------------------------------------------
* notvatel.c : novatel receiver dependent functions
*
*          Copyright (C) 2007,2008 by T.TAKASU, All rights reserved.
*
* version : $Revision: 1.2 $ $Date: 2008/07/14 00:05:05 $
* history : 2007/10/08 1.0 new
*           2008/05/09 1.1 fix bug lli flag outage
*           2008/06/16 1.2 separate common functions to rcvcmn.c
*           2009/04/01 1.3 add prn number check for raw obs data
*-----------------------------------------------------------------------------*/
#include "rtklib.h"
#include "novatel.h"

#define NTOBS       8
#define NOVHLEN     28          /* novatel message header length */
#define NOVSYN      0xAA4412    /* novatel message start sync code */
#define RANGECMP    140         /* message id : range compressed */
#define RAWEPH      41          /* message id : raw ephemeris */
#define IONUTC      8           /* message id : iono and utc information */
#define RAWWAASFRM  287         /* message id : raw waas frame */
#define U1(p)       (*((unsigned char *)(p)))
#define I1(p)       (*((char *)(p)))
#define U2(p)       (*((unsigned short *)(p)))
#define I2(p)       (*((short *)(p)))
#define U4(p)       (*((unsigned int *)(p)))
#define I4(p)       (*((int *)(p)))
#define R8(p)       (*((double *)(p)))
#define R4(p)       (*((float *)(p)))
#define WL1         0.1902936727984
#define WL2         0.2442102134246
#define POLYCRC32   0xEDB88320U /* CRC32 polynomial */

static const char rcsid[]="$Id: novatel.c,v 1.2 2008/07/14 00:05:05 TTAKA Exp $";

/* output rinex obs types */
static char tobs[NTOBS][3]={"C1","L1","D1","S1","P2","L2","D2","S2"};

/* application defined functions ---------------------------------------------*/
extern int showmsg(char *format,...);   /* show message and check abort */

/* extract unsigned bits from message ----------------------------------------*/
static unsigned int getbitu(const unsigned char *buff, int pos, int len)
{
    unsigned int bit=0;
    int i;
    for (i=pos;i<pos+len;i++) bit=(bit<<1)+((buff[i/8]>>(7-i%8))&1U);
    return bit;
}
/* extend sign ---------------------------------------------------------------*/
static int exsign(unsigned int v, int bits)
{
    return (int)(v&(1<<(bits-1))?v|(~0U<<bits):v);
}
/* CRC-32 parity -------------------------------------------------------------*/
static unsigned int crc32(const unsigned char *buff, int len)
{
    int i,j;
    unsigned int crc=0;
    for (i=0;i<len;i++) {
        crc^=buff[i];
        for (j=0;j<8;j++) {
            if (crc&1) crc=(crc>>1)^POLYCRC32; else crc>>=1;
        }
    }
    return crc;
}
/* decode tracking status ----------------------------------------------------*/
static int decodestat(unsigned int stat, int *track, int *plock, int *clock,
                      int *parity, int *halfc)
{
    *track=stat&0x1F;       /* tracking state */
                            /* (0=L1 idle, 1=L1 sky search, */
                            /*  2=L1 wide freq pull-in, 3=L1 narrow freq pull-in, */
                            /*  4=L1 pll, 5=L1 reacq, 6=L1 steering, 7=L1 fll, */
                            /*  8=L2 idle, 9=L2 p-code align, 10=L2 search, */
                            /*  11=L2 pll, 12=L2 steering) */
    *plock=(stat>>10)&1;    /* phase-lock flag (0=not locked, 1=locked) */
    *clock=(stat>>12)&1;    /* code-lock flag (0=not locked, 1=locked) */
    *parity=(stat>>11)&1;   /* parity known flag (0=not known, 1=known) */
    *halfc=(stat>>28)&1;    /* phase meas (0=half-cycle not added, 1=added) */
    return (stat>>21)&0x1F; /* signal type */
                            /* (GPS    : 0=L1C/A, 5=L2P, 9=L2Pcodeless, 17=L2C */
                            /*  SBAS   : 0=L1C/A */
                            /*  GLONASS: 0=L1C/A, 4=L1P, 5=L2P) */
}
/* decode RANGECMPB ----------------------------------------------------------*/
static int decoderangecmp(raw_t *raw, gtime_t time, const unsigned char *buff,
                          int len, int opt, obsd_t *obs)
{
    const double maxval=8388608.0;
    unsigned int lockc;
    double psr,adr,adr_rolls,tt;
    float dop;
    int i,j,n,m,nobs,sat,type,track,plock,clock,parity,halfc;
    unsigned char snr,lli;
    
    nobs=U4(buff);
    tt=timediff(time,raw->time);
    
    for (i=n=0,j=4;i<nobs&&n<MAXOBS;i++,j+=24) {
        type=decodestat(U4(buff+j),&track,&plock,&clock,&parity,&halfc);
        
        dop=(float)(exsign(U4(buff+j+4)&0xFFFFFFF,28)/256.0);
        psr=(U4(buff+j+7)>>4)/128.0+U1(buff+j+11)*2097152.0;
        adr=I4(buff+j+12)/256.0;
        adr_rolls=(psr/(type==0?WL1:WL2)+adr)/maxval;
        adr=-adr+maxval*floor(adr_rolls+(adr_rolls<=0?-0.5:0.5));
        sat=U1(buff+j+17);
        if (sat<=0||(MAXPRNGPS<sat&&sat<MINPRNSBS)||MAXPRNSBS<sat) continue;
        lockc=U4(buff+j+18)&0x1FFFFF; /* lock time (1/32 s) */
        
        lli=(int)(lockc-raw->lockc[SAT(sat)-1][type!=0]+1)/32.0<=tt||
            halfc!=raw->halfc[SAT(sat)-1][type!=0];
        raw->lockc[SAT(sat)-1][type!=0]=lockc;
        raw->halfc[SAT(sat)-1][type!=0]=halfc;
        if (!parity) lli|=2;
        
        snr=(unsigned char)(((U2(buff+j+20)&0x3FF)>>5)+20);
        if (!clock) psr=0.0;
        if (!plock) adr=dop=0.0;
        
        if (type==0) { /* L1 */
            obs[n].time=time;
            obs[n].sat=sat;
            obs[n].L[0]=adr; obs[n].L[1]=0.0;
            obs[n].P[0]=psr; obs[n].P[1]=0.0;
            obs[n].D[0]=dop; obs[n].D[1]=0.0;
            obs[n].SNR[0]=snr; obs[n].SNR[1]=0;
            obs[n].LLI[0]=lli; obs[n].LLI[1]=0;
            n++;
        }
        else { /* L2 */
            for (m=0;m<n;m++) if (obs[m].sat==sat) break;
            obs[m].L[1]=adr+(type==17?0.5:0); /* for L2C Phase shift */
            obs[m].P[1]=psr;
            obs[m].D[1]=dop;
            obs[m].SNR[1]=snr;
            obs[m].LLI[1]=lli;
            if (m>n) n=m;
        }
    }
    raw->time=time;
    return n;
}
/* decode RAWEPHB ------------------------------------------------------------*/
static int decoderaweph(gtime_t time, const unsigned char *buff, int len,
                        eph_t *ephs)
{
    eph_t eph={0};
    double ttr,ion[8],utc[4];
    unsigned int words[10];
    int i,j,sat,week,leaps,iode1=-1,iode2=-1;
    
    sat=U4(buff);
    if (len!=102||sat<=0||32<sat) return 0;
    eph.sat=(unsigned char)sat;
    eph.ttr=time;
    for (i=0;i<3;i++) {
        for (j=0;j<10;j++) words[j]=getbitu(buff+12+i*30,j*24,24);
        if (decodefrm(words,&eph,ion,utc,&leaps)!=i+1) return 0;
        if      (i==0) iode1=eph.iodc&0xFF;
        else if (i==1) iode2=eph.iode;
        else if (eph.iode!=iode1||eph.iode!=iode2) return 0;
    }
    ttr=time2gpst(eph.ttr,&week);
    if (ttr>302400.0&&eph.toes<302400.0) {
        eph.ttr=gpst2time(++week,ttr-604800.0);
    }
    eph.toe=gpst2time(week,eph.toes);
    if (ephs[sat-1].iode==eph.iode) return 0; /* duplicated */
    ephs[sat-1]=eph;
    return sat;
}
/* decode IONUTCB ------------------------------------------------------------*/
static int decodeionutc(gtime_t time, const unsigned char *buff, int len,
                        double *ion, double *utc, int *leaps)
{
    int i;
    
    if (len!=108) return 0;
    for (i=0;i<8;i++) ion[i]=R8(buff+i*8);
    utc[0]=R8(buff+72);
    utc[1]=R8(buff+80);
    utc[2]=(double)U4(buff+68);
    utc[3]=(double)U4(buff+64);
    *leaps=I4(buff+96);
    return 1;
}
/* decode RAWWAASFRMB --------------------------------------------------------*/
static int decoderawwaasfrm(gtime_t time, const unsigned char *buff, int len,
                            sbsmsg_t *sbsmsg)
{
    int i,sat;
    
    sat=U4(buff+4);
    if (len!=48||sat<MINPRNSBS||MAXPRNSBS<sat) return 0;
    sbsmsg->tow=(int)time2gpst(time,&sbsmsg->week);
    sbsmsg->sat=sat;
    for (i=0;i<29;i++) sbsmsg->msg[i]=buff[12+i];
    return sat;
}
/* decode novatel message ------------------------------------------------------
* decode and input a novatel message. support RANGECMPB, RAWEPHB, IONUTCB and
* RAWWAASFRMB.
* args   : raw_t *raw    IO     raw data control struct
*          unsigned char *buff I message buffer
*          int   len     I      buffer length (bytes)
*          int   opt     I      decode option
*          obs_t *obs    IO     observation data (NULL: no output)
*          nav_t *nav    IO     navigation message (NULL: no output)
*          sbsmsg_t *sbsmsg IO  sbas message (NULL: no output)
*          int   *sat    O      satellite prn number for ephemeris or sbas msg
* return : input status (-1:error, 0:no valid data, 1:obs data, 2:ephemeris,
*                        3:sbas message, 9:ion/utc parameters)
*-----------------------------------------------------------------------------*/
extern int decodenov(raw_t *raw, const unsigned char *buff, int len, int opt,
                     obs_t *obs, nav_t *nav, sbsmsg_t *sbsmsg, int *sat)
{
    gtime_t time;
    int hlen,msg,type;
    
    if (crc32(buff,len-4)!=U4(buff+len-4)) return -1; /* check crc */
    
    trace(3,"decodenov:msg=%2d\n",U2(buff+4));
    
    hlen=U1(buff+3);
    msg =U2(buff+4);
    type=U1(buff+6); /* 1:ascii,2:binary */
    time=gpst2time(U2(buff+14),U4(buff+16)*0.001);
    if (type!=2) return 0;
    
    switch (msg) {
    case RANGECMP:
        if (!obs||!screent(time,raw->ts,raw->te,raw->tint)) return 0;
        if ((obs->n=decoderangecmp(raw,time,buff+hlen,len-hlen-4,opt,obs->data))>0) {
            return 1;
        }
        break;
    case RAWEPH:
        if (nav&&(*sat=decoderaweph(time,buff+hlen,len-hlen-4,nav->eph))>0) {
            return screent(nav->eph[*sat-1].ttr,raw->ts,raw->te,0.0)?2:0;
        }
        break;
    case RAWWAASFRM:
        if (sbsmsg&&(*sat=decoderawwaasfrm(time,buff+hlen,len-hlen-4,sbsmsg))>0) {
            return screent(gpst2time(sbsmsg->week,sbsmsg->tow),raw->ts,raw->te,0.0)?3:0;
        }
        break;
    case IONUTC:
        if (nav&&decodeionutc(time,buff+hlen,len-hlen-4,nav->ion,nav->utc,&nav->leaps)) {
            return 9;
        }
        break;
    }
    return 0;
}
/* convert novatel log file ----------------------------------------------------
* convert navatel log file to rinex obs/nav and sbas message file. log file
* shall contain RANGECMPB, RAWEPHB, IONUTCB and RAWWAASFRMB.
* args   : FILE  *fp     I      input log file pointer
*          FILE  *ofp    I      output rinex obs file pointer (NULL:no output)
*          FILE  *nfp    I      output rinex nav file pointer (NULL:no output)
*          FILE  *sfp    I      output sbas message file pointer (NULL:no output)
*          gtime_t ts    I      time start
*          gtime_t te    I      time end
*          double tint   I      time interval
*          rnxopt_t *rnxopt I   rinex options
*          int   opt     I      conversion options
* return : none
*-----------------------------------------------------------------------------*/
extern void convnov(FILE *fp, FILE *ofp, FILE *nfp, FILE *sfp, gtime_t ts,
                    gtime_t te, double tint, const rnxopt_t *rnxopt, int opt)
{
    obsd_t data[MAXOBS]={{{0}}};
    eph_t eph[32]={{0}};
    obs_t obs={0};
    nav_t nav={0};
    sbsmsg_t sbsmsg;
    raw_t raw={{0}};
    gtime_t time={0},tstart={0};
    int i,c,sat,type,nmsg[4]={0};
    unsigned int sync=0;
    unsigned short len;
    char s[128]="",s1[64],s2[64];
    unsigned char buff[4096];
    
    obs.data=data; nav.eph=eph;
    for (i=0;i<32;i++) eph[i].iode=-1;
    raw.ts=ts; raw.te=te; raw.tint=tint;
    
    while ((c=fgetc(fp))!=EOF) {
        if (((sync=(sync<<8)+(unsigned char)c)&0xFFFFFF)!=NOVSYN) continue;
        for (i=0;i<3;i++) buff[i]=(unsigned char)(NOVSYN>>(2-i)*8);
        if (fread(buff+3,NOVHLEN-3,1,fp)<1) break;
        len=((unsigned short)buff[9]<<8)+buff[8];
        if (NOVHLEN+len+4>sizeof(buff)) {nmsg[3]++; continue;}
        if ((int)fread(buff+NOVHLEN,1,len+4,fp)<len) break;
        
        type=decodenov(&raw,buff,NOVHLEN+len+4,opt,&obs,&nav,&sbsmsg,&sat);
        
        if (type==1&&ofp) {
            if (nmsg[0]==0) {
                outrnxobsh(ofp,rnxopt,tobs,NTOBS,obs.data[0].time);
            }
            outrnxobsb(ofp,obs.data,obs.n,0,tobs,NTOBS);
            nmsg[0]++;
            time=obs.data[0].time;
        }
        else if (type==2&&nfp) {
            if (nmsg[1]==0) {
                outrnxnavh(nfp,rnxopt,nav.ion,nav.utc,nav.leaps);
            }
            outrnxnavb(nfp,&nav.eph[sat-1]);
            nmsg[1]++;
            if (!ofp) time=nav.eph[sat-1].ttr;
        }
        else if (type==3&&sfp) {
            sbsoutmsg(sfp,&sbsmsg);
            nmsg[2]++;
            if (!ofp&&!nfp) time=gpst2time(sbsmsg.week,sbsmsg.tow);
        }
        else if (type==-1) nmsg[3]++;
        if (te.time!=0&&timediff(time,te)>DTTOL+1.0) break;
        if (time.time!=0) {
            time2str(time,s1,0); sprintf(s,"%s: ",s1);
            if (tstart.time==0) tstart=time;
        }
        if (showmsg("%sObs=%d Nav=%d Sbs=%d Err=%d",s,nmsg[0],nmsg[1],
                    nmsg[2],nmsg[3])) break;
    }
    if (nfp) {
        rewind(nfp);
        outrnxnavh(nfp,rnxopt,nav.ion,nav.utc,nav.leaps);
    }
    if (tstart.time!=0) {
        time2str(tstart,s1,0); time2str(time,s2,0);
        sprintf(s,"%s-%s: ",s1,s2+5);
    }
    showmsg("%sObs=%d Nav=%d Sbs=%d Err=%d",s,nmsg[0],nmsg[1],nmsg[2],nmsg[3]);
}
/* read novatel log file -------------------------------------------------------
* read novatel log file. input log file shall contain RANGECMPB, RAWEPHB,
* IONUTCB and RAWWAASFRMB
* args   : FILE  *fp     I      input file pointer
*          gtime_t ts,te I      time start/end
*          double tint   I      time interval
*          int   opt     I      read option)
*          obs_t *obs    IO     observation data    (NULL: no output)
*          nav_t *nav    IO     navigation messages (NULL: no output)
*          sbs_t *sbs    IO     sbas messages       (NULL: no output)
*          int   *nmax   I      max number of messages 
* return : none
*-----------------------------------------------------------------------------*/
extern void readnov(FILE *fp, gtime_t ts, gtime_t te, double tint, int opt,
                    obs_t *obs, nav_t *nav, sbs_t *sbs, int *nmax)
{
    obsd_t data[MAXOBS]={{{0}}};
    eph_t eph[32]={{0}};
    obs_t obsp={0};
    nav_t navp={0};
    sbsmsg_t sbsp;
    raw_t raw={{0}};
    int i,c,sat,type;
    unsigned int sync=0;
    unsigned short len;
    unsigned char buff[4096];
    
    obsp.data=data; navp.eph=eph;
    for (i=0;i<32;i++) eph[i].iode=-1;
    raw.ts=ts; raw.te=te; raw.tint=tint;
    
    while ((c=fgetc(fp))!=EOF) {
        if (((sync=(sync<<8)+(unsigned char)c)&0xFFFFFF)!=NOVSYN) continue;
        for (i=0;i<3;i++) buff[i]=(unsigned char)(NOVSYN>>(2-i)*8);
        if (fread(buff+3,NOVHLEN-3,1,fp)<1) break;
        len=((unsigned short)buff[9]<<8)+buff[8];
        if (NOVHLEN+len+4>sizeof(buff)) continue;
        if ((int)fread(buff+NOVHLEN,1,len+4,fp)<len) break;
        
        type=decodenov(&raw,buff,NOVHLEN+len+4,opt,&obsp,&navp,&sbsp,&sat);
        
        if (type==1&&obs) {
            addobs(obs,&obsp,nmax);
        }
        else if (type==2&&nav) {
            addnav(nav,&navp,sat,nmax+1);
        }
        else if (type==3&&sbs) {
            addsbs(sbs,sbsp,nmax+2);
        }
        else if (type==9&&nav) {
            addionutc(nav,&navp);
        }
        if (te.time!=0&&timediff(obsp.data[0].time,te)>DTTOL) break;
    }
}
